program D80toTAP;       {Program pro prevod souboru z .D80 do .TAP}
                        {Funguje pro diskove obrazy formatu 80x9 (D80)}

                        {verze 1.1 01.09.2002 22:25 X-agon of Phantasy}
                        {TODO: -prepis existujiciho souboru}
                        {      -problem dlouhych jmen      }

uses crt, dos;

const

	SMAZANY                   = 229;	{hodnota v adresari indikujici
                                        	smazany soubor}

	D80velikost_sektoru       = 512;	{v bytech}

	D80boot_start             = 0;          {pocatecni sektor}
       	D80boot_sektor            = 1;          {velikost v sektorech}
        D80FAT_start              = 1;		{pocatecni sektor}
	D80FAT_velikost		  = 5;          {velikost v sektorech}
        D80adresar_start          = 6;		{pocatecni sektor}
        D80adresar_velikost       = 8;          {velikost v sektorech}
        D80adresar_delka_polozky  = 32;         {v bytech}
        D80adresar_max_souboru    = 128;        {max. pocet souboru v adresari}
        D80data_start             = 14;		{pocatecni sektor}
        D80data                   = 1426;       {velikost v sektorech}

	{polozky FAT jsou 1,5Bajtove (12bitove) hodnoty}

	D80FAT_pocet_polozek      = 1705;
	D80FAT_polozek_na_sektor  = 341;
	D80FAT_koncova_znacka_id  = 3584;	{12bitova hodnota}

	{FAT: Znacka konce souboru je vytvorena takto: delka souboru je vydelena
	512 a ke zbytku po deleni (0-511) pricteme 3584 a to je ono. Je zrejme,
	ze toto neplati pro prazdne soubory.}

	D80max_delka_souboru      = 129;	{128 sektoru + 1 jako zarazka}

	TAPdelka_hlavicky         = 17;		{v bytech}
	TAPdelka_nazvu_souboru    = 10;		{v bytech}


type zsoubor = record
                jmeno: string[12];	{10 znaku + '.' + pripona (PNCB)}
                start_adresa: word;
                delka: word;
		start_sektor: word;
	        basic_promenne: word;
        end;

var

        fat:      array[1..D80FAT_pocet_polozek] of word;
        adresar:  array[1..D80adresar_velikost*D80velikost_sektoru] of byte;
	soubory:  array[1..D80adresar_max_souboru] of zsoubor;
        sektor:   array[1..D80velikost_sektoru] of byte;
	hlavicka: array[1..TAPdelka_hlavicky] of byte;
	stezka:   array[1..D80max_delka_souboru] of word;

	D: DirStr;
	N: NameStr;
	E: ExtStr;

        znak, pripona: char;
        bajt, iter, pocet_souboru, checksum: byte;
	pocet_stop_na_strane, pocet_sektoru_na_stope: byte;
	status: shortint;
	slovo, slovo2: word;
	offset: longint;
	vstupni_soubor, vystupni_soubor: file;
	nazev_souboru, nazev_vystupniho_souboru, nazev_disku, d80fname: string;


{**********************************************}
{  Pomocne funkce  }

function GetFname(var cislo_souboru: byte): string;
{vraci string s nazvem souboru a priponou, cislo_souboru je mezi 0 az 127}
{vraci prazdny retezec pri smazanem souboru/chybnem parametru cislo_souboru }
var
	pripona, pom: byte;
        i, offset: word;
	temp: string;
	znak: char;

begin
	temp := '';

	if ((cislo_souboru < 0) or (cislo_souboru > 127)) then
	begin
        	GetFname := '';
	        exit;
	end;

	offset := cislo_souboru * D80adresar_delka_polozky + 1;
	pripona := adresar[offset];

	if (pripona = SMAZANY) then
	begin
		GetFname := '';
                exit;
	end;

	for i := 1 to 10 do
        begin
        	pom := adresar[offset+i];

        	if (pom <> 0) then temp := temp + chr(pom)
        	              else temp := temp + ' ';

	end;
        temp := temp + '.' + chr(pripona);

	GetFname := temp;
end;


function GetLength(var cislo_souboru: byte): word;
{Vraci delku souboru}
var  low, hi: longint;

begin
	low := cislo_souboru*D80adresar_delka_polozky + 1 + 11;
	hi := low + 1;

	GetLength := adresar[low] + 256*adresar[hi];
end;


function GetStartAddress(var cislo_souboru: byte): word;
{Vraci startovaci adresu (u CODE, u BASICu vraci start. radek)}
var  
	low, hi: longint;

begin
	low := cislo_souboru*D80adresar_delka_polozky + 1 + 13;
	hi := low + 1;

	GetStartAddress := adresar[low] + 256*adresar[hi];
end;


function GetStartSector(var cislo_souboru: byte): word;
{Vraci prvni sektor souboru (vstupni bod do FAT)}
var  
	low, hi: longint;

begin
	low := cislo_souboru*D80adresar_delka_polozky + 1 + 17;
	hi := low + 1;

	GetStartSector := adresar[low] + 256*adresar[hi];
end;

function GetBasicVariables(var cislo_souboru: byte): word;
{Vraci hodnotu z hlavicky 'delka BASIC bloku bez promennych}
var  
	low, hi: longint;

begin
	low := cislo_souboru*D80adresar_delka_polozky + 1 + 15;
	hi := low + 1;

	GetBasicVariables := adresar[low] + 256*adresar[hi];
end;


function GetDiskParams:shortint;
{Zjisti parametry diskety a vrati nulu, jde-li o MDOS disketu, jinak vraci -1.}

var iter: byte;

begin

        {Zjisti jmeno diskety}
        seek(vstupni_soubor, 0);
        blockread(vstupni_soubor, sektor, 1);

        nazev_disku := '';
        for iter:= 1 to 10 do
        begin
        	if (sektor[191+iter] <> 0) then 
		nazev_disku := nazev_disku + chr(sektor[191+iter]);
        end;

        {Zjisti pocet stop na strane}
	pocet_stop_na_strane := sektor[179];

        {Zjisti pocet sektoru na stopu}
	pocet_sektoru_na_stope:= sektor[180];


	{V bootu na adresach 204 - 207 lezi retezec 'SDOS'}
	if (chr(sektor[205]) = 'S')   then GetDiskParams:= 0
				      else GetDiskParams:= -1;
end;



{**********************************************}
{  Pomocne procedury  }


procedure GetFilesInfo;
{Naplni pole zaznamu o souborech 'soubory'}

var iter: byte;

begin
        pocet_souboru := 0;

        seek(vstupni_soubor, D80adresar_start);
        blockread(vstupni_soubor, adresar, D80adresar_velikost);

        for iter := 0 to D80adresar_max_souboru - 1 do
        begin
                d80fname := GetFname(iter);

                if (d80fname <> '') then
		begin
                	pocet_souboru := pocet_souboru + 1;

                	with soubory[pocet_souboru] do
  			begin
                    		jmeno:= d80fname;
                    		start_adresa:= GetStartAddress(iter);
                    		delka:= GetLength(iter);
                    		start_sektor:= GetStartSector(iter);
                    		basic_promenne:= GetBasicVariables(iter);
               		end;
        	end;
        end;
end;


procedure DisplayDiskContent;
{Zobraz nazvy souboru, delky, pocatecni sektor}
var iter: byte;

begin
	writeln('          DiskName: ', nazev_disku);
	writeln('Ŀ');
	writeln('Nr.  Filename  StartLengthSector');
	writeln('Ĵ');
        
	for iter:=1 to pocet_souboru do
        begin
                with soubory[iter] do
                writeln('',iter:3,'',jmeno,'',start_adresa:5,'',delka:6,'',start_sektor:6,'');
        end;
	writeln('');
end;


procedure FillFATTable(cislo_FAT_polozky: word);
{Procedura vezme pole 'sektor' a vytahne z nej FAT polozky a ulozi
 je do pole 'fat', podle 'cislo_fat_polozky'}
var 
	hodnota1, hodnota2, iter: word;
	byte2, low1, low2, hi: byte;

begin
        iter:=1;
	while (iter <= D80velikost_sektoru - 2 ) do
        begin
                low1  := sektor[iter];
        	byte2 := sektor[iter+1];
	        low2  := sektor[iter+2];

        	hi := byte2 and 240; {%11110000}
	        hi := hi shr 4;      {presun do dolniho pulbajtu}
                hodnota1 := 256*hi + low1;

        	hi := byte2 and 15;  {%00001111}
	        hodnota2 := 256*hi + low2;

	        fat[cislo_FAT_polozky] := hodnota1;
	        fat[cislo_FAT_polozky+1] := hodnota2;

	        cislo_FAT_polozky := cislo_FAT_polozky + 2;

	        iter := iter + 3;
	end;

        {Pridame posledni polozku ze "zalomeni FATky" (2bajty, nizsi
         pulbyte druheho bajtu ma vzdy hodnotu 13 (%1101)}

        	low1 := sektor[iter];
	        hi := sektor[iter+1] and 240; {%11110000}
	        hi := hi shr 4;               {presun do dolniho pulbajtu}
	        hodnota1 := 256*hi + low1;

	        fat[cislo_FAT_polozky] := hodnota1;

end;


procedure LoadFATTable;
{Prevede 1,5Bajtove hodnoty na 2bajtove (kvuli snadnejsimu pristupu) a
 ulozi vysledek do pole 'fat'}
var iter: byte;

begin
        for iter:= 0 to 4 do
	begin
        	seek(vstupni_soubor, iter+D80FAT_start);
        	blockread(vstupni_soubor, sektor, 1);

		FillFATTable(iter*D80FAT_polozek_na_sektor+1); {+1 - index v poli}
	end;
end;


procedure LoadFileTrack(cislo_souboru:byte);
{Do pole 'stezka' ulozi posloupnost sektoru, kterymi prochazi soubor
 'cislo_souboru'. 'cislo_souboru ukazuje do pole 'soubory'. Pri chybe
 neudela nic.}
var 
	iter, akt_sektor, koncova_znacka_ve_FAT: word;
	delka_souboru: word;
	pom: byte;
	nazev_souboru: string;

begin
	for iter:= 1 to D80max_delka_souboru do stezka[iter]:= 0;

        if (cislo_souboru < 1) or (cislo_souboru > pocet_souboru) then exit;

        nazev_souboru := soubory[cislo_souboru].jmeno;
	delka_souboru := soubory[cislo_souboru].delka;
        akt_sektor := soubory[cislo_souboru].start_sektor;
	koncova_znacka_ve_FAT := delka_souboru mod 512 + D80FAT_koncova_znacka_id;

	iter:= 1; {index do pole 'stezka'}

        if (delka_souboru = 0) then exit; {stezka[1] := 0 -> prazdny soubor}

	while (akt_sektor <> koncova_znacka_ve_FAT) do
	begin
        	stezka[iter]:= akt_sektor;
	        akt_sektor := fat[akt_sektor+1];
	        inc(iter);
	end;
end;

function CutExtension(nazev_souboru: string):string;
{Odsekne vse napravo od posledni '.' v nazvu souboru}
var
        iter, znacka: byte;
	pom: string;
begin
        pom := '';
        iter:= length(nazev_souboru);
	while (nazev_souboru[iter] <> '.') do dec(iter);
	for znacka:=1 to iter do pom:= pom + nazev_souboru[znacka];
	pom := pom + 'tap';

	CutExtension := pom;
	nazev_vystupniho_souboru:= pom;
end;

procedure CreateAndSaveHeader(cislo_souboru: byte);
{Procedura vytvori korektni hlavicku a ulozi ji do souboru}
var
        checksum, i, prip_byte: byte;
	pripona: char;

	uvod_delka: word;
	flag: byte;

begin

	checksum := 0; uvod_delka:= 19; flag:=0;

        for i:=1 to sizeof(hlavicka) do hlavicka[i]:=0;

	pripona := soubory[cislo_souboru].jmeno[12];

        case pripona of
          'P':  prip_byte := 0;
          'N':  prip_byte := 1;
          'C':  prip_byte := 2;
          'B':  prip_byte := 3;
        else    prip_byte := 3; {Na disku mohou byt i jine pripony, budou CODE}
        end;

	{typ souboru - PCNB}
	hlavicka[1] := prip_byte;

	{10 znaku jmeno souboru}
	for i:=1 to TAPdelka_nazvu_souboru do
        hlavicka[1+i] := ord(soubory[cislo_souboru].jmeno[i]);

	{2 bajty delka souboru}
        hlavicka[12] := lo(soubory[cislo_souboru].delka);
	hlavicka[13] := hi(soubory[cislo_souboru].delka);

	{2 bajty pocatecni adresa (nebo Basic autostart)}
        hlavicka[14] := lo(soubory[cislo_souboru].start_adresa);
	hlavicka[15] := hi(soubory[cislo_souboru].start_adresa);

        {2 bajty delka Basicu bez promennych (nebo 32768 pro CODE)}
        hlavicka[16] := lo(soubory[cislo_souboru].basic_promenne);
	hlavicka[17] := hi(soubory[cislo_souboru].basic_promenne);

	{Uvodni nula - flag - nema na checksum vliv}
	for i:=1 to sizeof(hlavicka) do checksum := checksum xor hlavicka[i];

	blockwrite(vystupni_soubor, uvod_delka, sizeof(uvod_delka));
        blockwrite(vystupni_soubor, flag, sizeof(flag));
	blockwrite(vystupni_soubor, hlavicka, sizeof(hlavicka));
	blockwrite(vystupni_soubor, checksum, sizeof(checksum));

end;


procedure CreateAndSaveBody(cislo_souboru: byte);
{Procedura ulozi blok + checksum do souboru}
var
        checksum, jter, pocet_sektoru: byte;
	uvod_delka, zbytek, nula, itersek: word;
	offset_v_souboru: longint;
	flag: byte;

begin

        checksum:= 0; flag:= 255; checksum:= checksum xor flag;
        {vynulujeme prislusne promenne}

        {Prazdne soubory osetrime hned na zacatku}
        if (soubory[cislo_souboru].delka = 0) then {nulova delka = 02 00 FF FF }
        begin                                      {               ^^^^^ ^^ ^^ }
            nula:= $0002;                          {               delka fl crc}
            blockwrite(vystupni_soubor, nula, sizeof(nula));
            nula:= $FFFF;
            blockwrite(vystupni_soubor, nula, sizeof(nula));
            exit;
        end;

        uvod_delka:= soubory[cislo_souboru].delka + sizeof(flag) + sizeof(checksum);
        {prvni 2 bytes obsahuji delku dat+checksum+flag}

        blockwrite(vystupni_soubor, uvod_delka, sizeof(uvod_delka));
        {ulozime na disk}
        blockwrite(vystupni_soubor, flag, sizeof(flag));
        {1byte flag, ted nasleduji samotna data}

        LoadFileTrack(cislo_souboru);

        jter := 1; pocet_sektoru:= 0;
        while(stezka[jter] <> 0) do
        begin
                inc(jter);
		inc(pocet_sektoru);
        end;

        zbytek := soubory[cislo_souboru].delka mod D80velikost_sektoru;
        {kolik bytes ulozit z posledniho sektoru}

        jter:= 1;
        while (jter < pocet_sektoru) do {az do pocet_sektoru - 1}
        begin
                offset_v_souboru:= longint(stezka[jter]);
                seek(vstupni_soubor, offset_v_souboru);
		{stezka[iter] obsahuje cislo sektoru, ktery budeme ukladat}
		blockread(vstupni_soubor, sektor, 1);
		{ulozime jej do pole o velikosti 512 bytes}

		blockwrite(vystupni_soubor, sektor, sizeof(sektor));
		{uloz sektor do souboru}

		for itersek:=1 to sizeof(sektor) do checksum:= checksum xor sektor[itersek];
		{a vypocitej checksum (mezivypocet)}

		inc(jter);
        end;

                offset_v_souboru:= longint(stezka[jter]);
		seek(vstupni_soubor, offset_v_souboru);

		{posledni sektor osetrime zvlast}
		blockread(vstupni_soubor, sektor, 1);
		{opet ulozime}

		if (zbytek = 0) then {delka souboru je delitelna 512}
		begin
			blockwrite(vystupni_soubor, sektor, sizeof(sektor));
			for itersek:=1 to sizeof(sektor) do 
			checksum:= checksum xor sektor[itersek];
		end
		else
		begin
			blockwrite(vystupni_soubor, sektor, zbytek);
			for itersek:=1 to zbytek do 
			checksum:= checksum xor sektor[itersek];
		end;
		{mame ulozeny vsechny bloky}

		blockwrite(vystupni_soubor, checksum, sizeof(checksum));
		{a nakonec ulozit kontrolni soucet}

end;


procedure MakeTapFile;
{Vezme vsechny informace a vyplazne .TAP soubor :-)}
var
        cislo_souboru: byte;
	pom_nazev: string;

begin
        pom_nazev := CutExtension(nazev_souboru);
	assign(vystupni_soubor, pom_nazev);
	{$I-}
	rewrite(vystupni_soubor, 1);
	{$I+}

	if IOResult <> 0 then
	begin
        	close(vstupni_soubor);
	        writeln('Soubor "',pom_nazev, '" se nepodarilo otevrit pro zapis!');
        	close(Output);
	        halt;
 	end;


        for cislo_souboru:= 1 to pocet_souboru do
	begin
        	CreateAndSaveHeader(cislo_souboru);
	        CreateAndSaveBody(cislo_souboru);
	end;

	close(vystupni_soubor);
end;


{*********************************************}
            { HLAVNI PROGRAM }
{*********************************************}

begin

  clrscr;
  textcolor(LightGray);
  textbackground(Black);

        {Budeme chtit presmerovat vystup do souboru}
        assign(Output,'');
        rewrite(Output);

        Fsplit(ParamStr(0), D, N, E);

        if (ParamCount <> 1) then
	begin
            writeln('ZX Spectrum .D80 => .TAP conversion utility');
            writeln;
            write('Converts .D80 image files created using RealSpec emulator');
            writeln(' to .TAP');
            writeln;
            writeln('Usage: ', N, ' image.d80');
            writeln;
            writeln('Send bugreports and comments to xhotarek@fi.muni.cz');
            writeln('Homepage: http://www.fi.muni.cz/~xhotarek');
            writeln;
            writeln('(c) 2002-08-31 X-agon of Phantasy');

            halt;
        end;


        nazev_souboru := ParamStr(1);

        assign(vstupni_soubor, nazev_souboru);
        {$I-}
        reset(vstupni_soubor, D80velikost_sektoru);
        {$I+}

                if IOResult <> 0 then
                begin
	                writeln('File "', nazev_souboru, '" not found!');
			close(Output);
                	halt;
		end;

	{Zjisti jmeno diskety, pocet sektoru na stopu a proved test na D80 disk}
	status:= GetDiskParams;

        if (status = -1) then
        begin
                close(vstupni_soubor);
                writeln('File "', nazev_souboru, '" is not a valid .D80 image!');
                close(Output);
		halt;
        end;

        {Napln pole 'soubory' infem o souborech na disku}
        GetFilesInfo;

        if (pocet_souboru = 0) then
        begin
        	close(vstupni_soubor);
                writeln('Image "',nazev_souboru,'" contains no files!');
		close(Output);
                halt;
        end;

	{Zobraz v tabulce informace o souborech}
	DisplayDiskContent;

        {Napln strukturu 'fat' 2bajtovymi hodnotami FAT polozek}
        LoadFATTable;

        {Udelej konecne to, k cemu je delany cely program :-)}
        MakeTapFile;

	writeln;
	writeln('Saved to "', nazev_vystupniho_souboru, '".');
	writeln('Done...');
	writeln;

	close(Output);

        close(vstupni_soubor);

end.
